/*
 * ARM920T Cirrus Logic EP9307 SPI Flash interface.
 *
 * Copyright (c) 2005 Ben Dubb, IPC Incorporated
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */


#include <common.h>
#include <linux/ctype.h>
#include <malloc.h>
#include <post.h>
#include <net.h>

#include <ep93xx.h>
#include <watchdog.h>
#include <linux/byteorder/swab.h>
//#include <spi.h>

#if defined(CONFIG_SPI)


#define SPI_EEPROM_WREN		     0x06
#define SPI_EEPROM_RDSR		     0x05
#define SPI_EEPROM_READ		     0x03
#define SPI_EEPROM_WRITE	     0x02
#define SPI_EEPROM_SECTOR_ERASE_ST     0xD8
#define SPI_EEPROM_SECTOR_ERASE_ATMEL  0x52

static int spi_rxclear(void);
static int spi_tx_wait(void);
static int spi_readbyte(void);
int spi_checkeeprom(void);
void inline spin_wheel (void);
ssize_t spi_erase (ulong addr);
int sflash_print_info(void);
unsigned long flash_init(void);


ssize_t word_spi_write (ulong addr, uchar *buffer, int len);
ssize_t word_spi_read (ulong addr, uchar *buffer, int len);


typedef struct {
    unsigned long SSPCR0;  /* control reg 0 */
    unsigned long SSPCR1;  /* control reg 1 */
    unsigned long SSPDR;   /* Tx/Rx data fifo reg */
    unsigned long SSPSR;   /* status reg */
    unsigned long SSPCPSR; /* clock prescale reg */
    unsigned long SSPICR;  /* interrupt status/clear reg */
} EP9307_SSP_T;

volatile EP9307_SSP_T *_ssp = (EP9307_SSP_T *)EP9307_SSP;

#define SSP_SET(reg, data) (_ssp->reg = data)
#define SSP_GET(reg)       (_ssp->reg)


unsigned int sflash_type = 0;
#define SFLASH_TYPE_ATMEL  (1 << 16)
#define SFLASH_TYPE_ST     (2 << 16)
#define SFLASH_TYPE_2M     18
#define SFLASH_TYPE_1M     17

//#define DEBUG

#ifdef	DEBUG
#define	DPRINT(a)	printf a;
#else	/* !DEBUG */
#define	DPRINT(a)
#endif	/* DEBUG */

/* -------------------
 * Function prototypes
 * ------------------- */
ssize_t spi_read (uchar *, int, uchar *, int);
ssize_t spi_write (uchar *, int, uchar *, int);


static int spi_tx_wait()
{
    while (1)
    {
        /* wait till Tx fifo empty */
        if (SSP_GET(SSPSR) & 0x01) 
            break;
    }

    while (1)
    {
        /* wait till not busy */        
        if (!(SSP_GET(SSPSR) & 0x10))
			break;
    }

    udelay(200);
    return 0;
}

static int spi_rxclear()
{
    int dummy;
  
    while (1)
    {
        /* wait till Rx fifo is empty */
        if ((SSP_GET(SSPSR) & 0x04))
            dummy = SSP_GET(SSPDR);
        else
            break;
    }
    return 0;
}

static int spi_readbyte()
{
	int byte;
    while (1)
    {
        /* wait till Rx fifo is non empty */
        if ((SSP_GET(SSPSR) & 0x04))
            break;
    }

	byte = SSP_GET(SSPDR);
    return (byte);
}


int spi_checkeeprom()
{
    unsigned int partid, mfgid;

    /* Is it ST/Windbound part */
    SSP_SET(SSPDR, 0xAB);                       /* Device ID */
    SSP_SET(SSPDR, 0);
    SSP_SET(SSPDR, 0);
    SSP_SET(SSPDR, 0);
    SSP_SET(SSPDR, 0); /* do one extra write to read the data */
    spi_tx_wait();

    spi_readbyte(); /* drop the 1st 4 bytes */
    spi_readbyte();
    spi_readbyte();
    spi_readbyte();
    partid = spi_readbyte();

    if (partid == 0x11)
    {
        sflash_type = SFLASH_TYPE_ST | SFLASH_TYPE_2M;
        return 1;
    }


    /* Is it Atmel AT25F2048? */
    SSP_SET(SSPDR, 0x15);                       /* Device ID */
    SSP_SET(SSPDR, 0);
    SSP_SET(SSPDR, 0);
    spi_tx_wait();

    spi_readbyte(); /* drop the 1st byte */
    mfgid = spi_readbyte();
    partid = spi_readbyte();
    if ((mfgid == 0x1f) && (partid == 0x63))
    {
        sflash_type = SFLASH_TYPE_ATMEL | SFLASH_TYPE_2M;
        return 1;
    }
    else if ((mfgid == 0x1f) && (partid == 0x60))
    {
        sflash_type = SFLASH_TYPE_ATMEL | SFLASH_TYPE_1M;
        return 1;
    }

    return 0;
}

unsigned long flash_init()
{
    long size = 0;
    
    //return 0;

    spi_init_f();
    
    spi_checkeeprom();
    if (sflash_type & 0x1f)
    {
        size = 1 << (sflash_type & 0x1f);
    }
    return (size);
}

int sflash_print_info()
{
    switch (sflash_type)
    {
    case SFLASH_TYPE_ATMEL | SFLASH_TYPE_1M:
        printf ("Atmel AT25F10");
        break;
    case SFLASH_TYPE_ATMEL | SFLASH_TYPE_2M:
        printf ("Atmel AT25F20");
        break;
    case SFLASH_TYPE_ST | SFLASH_TYPE_2M:
        printf ("ST W25P20");
        break;
    case SFLASH_TYPE_ST | SFLASH_TYPE_1M:
        printf ("ST W25P10");
        break;
    default:
        printf ("Unknown flash device");
    }

    return 0;
}

int spi_erase_flash_sector(ulong offset)
{
    return spi_erase(offset);
}


/* **************************************************************************
 *
 *  Function:    spi_init_f
 *
 *  Description: Init SPI-Controller (ROM part)
 *
 *  return:      ---
 *
 * *********************************************************************** */

void spi_init_f (void)
{
    SSP_SET(SSPCR0,  0x1c7);	/* program CR0 reg */
    SSP_SET(SSPCPSR, 2);		/* program predivisor reg */
    SSP_SET(SSPCR1,  0);	    /* deactivate SPI */
    SSP_SET(SSPCR1,  0x10);		/* activate SPI */

    spi_rxclear();

	return;
}


/****************************************************************************
 *  Function:    spi_write
 **************************************************************************** */
#define TRANSFER_LEN 4 /* one 32bit word at a time */

ssize_t spi_write (uchar *addr, int alen, uchar *buffer, int len)
{
	int index,i,newlen;
	int curraddr;

	curraddr=(addr[alen-3]<<16)+(addr[alen-2]<<8)+addr[alen-1];
	i=len;
	index=0;
	do {
		if(i>TRANSFER_LEN) {
			newlen=TRANSFER_LEN;
			i-=TRANSFER_LEN;
		}
		else {
			newlen=i;
			i=0;
		}
		word_spi_write (curraddr, &buffer[index], newlen);
		index+=newlen;
		curraddr+=newlen;
        if (index % 0x20) spin_wheel();
	}while(i);
    printf("\r");
	return (len);
}

/****************************************************************************
 *  Function:    spi_read
 **************************************************************************** */
ssize_t spi_read (uchar *addr, int alen, uchar *buffer, int len)
{
	int index,i,newlen;
	int curraddr;

	curraddr=(addr[alen-3]<<16)+(addr[alen-2]<<8)+addr[alen-1];

//	printf ("\nspi_read offset 0x%x, buffer 0x%x\n", curraddr, buffer);

	i=len;
	index=0;
	do {
		if(i>TRANSFER_LEN) {
			newlen=TRANSFER_LEN;
			i-=TRANSFER_LEN;
            WATCHDOG_RESET();
		}
		else {
			newlen=i;
			i=0;
		}
		word_spi_read (curraddr, &buffer[index], newlen);
		index+=newlen;
		curraddr+=newlen;
	}while(i);
	return (len);
}


ssize_t word_spi_write (ulong addr, uchar *buffer, int len)
{
	int i,dlen;
	i=0;

	//printf ("\nwrite word(%d) %x to offset 0x%x\n", len, *(ulong *)buffer, addr);
    udelay(1000);

    SSP_SET(SSPDR, SPI_EEPROM_WREN);            /* send write enable command */
    udelay(1000);
    spi_rxclear();
    SSP_SET(SSPDR, SPI_EEPROM_WRITE);           /* write memory  */
    SSP_SET(SSPDR, ((addr >> 16) & 0xff));
    SSP_SET(SSPDR, ((addr >> 8) & 0xff));
    SSP_SET(SSPDR, (addr & 0xff));

	for(dlen=0; dlen<len; dlen++) {
		SSP_SET(SSPDR, buffer[dlen]); /* WRITE memory array */
	}

    udelay(10000);

	/* ignore received data	*/
	spi_rxclear();
	for (i = 0; i < 1000; i++) {
		SSP_SET(SSPDR, SPI_EEPROM_RDSR); /* read status */
		if (!(SSP_GET(SSPDR) & 1)) {
			break;
		}
		udelay(1000);
	}
	spi_rxclear();
	if (i >= 1000) {
		printf ("*** spi_write: Time out while writing!\n");
	}
	return len;
}

ssize_t word_spi_read (ulong addr, uchar *buffer, int len)
{
	int i;
    int word;
    unsigned char *pBuf = (unsigned char *)&word;

	spi_rxclear();

    SSP_SET(SSPDR, SPI_EEPROM_READ);            /* send read command */
    SSP_SET(SSPDR, ((addr >> 16) & 0xff));      /* then the address */
    SSP_SET(SSPDR, ((addr >> 8) & 0xff));
    SSP_SET(SSPDR, (addr & 0xff));

    spi_readbyte();

    /* prime the read buffer */
    SSP_SET(SSPDR, 0);
    SSP_SET(SSPDR, 0);
    SSP_SET(SSPDR, 0);
    SSP_SET(SSPDR, 0);

    /* read and throw away the address bytes */
    spi_readbyte();
    spi_readbyte();
    spi_readbyte();

//    spi_tx_wait();

	for(i=0;i<len;i++) {
		*pBuf++=(char)spi_readbyte();
	}
    spi_rxclear();
	
//	printf("0x%08x ", word);
	*(unsigned long *)buffer = word;

	return len;
}

ssize_t spi_erase (ulong addr)
{
	//printf ("\nspi_erase offset 0x%x\n", addr);

    spi_rxclear();

    SSP_SET(SSPDR, SPI_EEPROM_WREN);            /* send write enable command */
    udelay(1000);
    spi_rxclear();

    if (sflash_type & SFLASH_TYPE_ST)
    {
        SSP_SET(SSPDR, SPI_EEPROM_SECTOR_ERASE_ST);    /* send sector erase command */
        SSP_SET(SSPDR, ((addr >> 16) & 0xff));         /* then the address */
        SSP_SET(SSPDR, ((addr >> 8) & 0xff));
        SSP_SET(SSPDR, (addr & 0xff));

        /* read and throw away the address bytes */
        spi_readbyte();
        spi_readbyte();
        spi_readbyte();
        spi_readbyte();
    }
    else if (sflash_type & SFLASH_TYPE_ATMEL)
    {
        SSP_SET(SSPDR, SPI_EEPROM_SECTOR_ERASE_ATMEL); /* send sector erase command */
        SSP_SET(SSPDR, ((addr >> 16) & 0xff));         /* then the address */
        SSP_SET(SSPDR, ((addr >> 8) & 0xff));
        SSP_SET(SSPDR, (addr & 0xff));

        /* read and throw away the address bytes */
        spi_readbyte();
        spi_readbyte();
        spi_readbyte();
        spi_readbyte();
    }
    else
    {
        return -1;
    }

    while (1)
    {
        int byte, i=0;

        //printf("read RDSR\n");
        SSP_SET(SSPDR, SPI_EEPROM_RDSR); /* read status */

        /* prime the fifo */
        SSP_SET(SSPDR, 0);
        spi_tx_wait();

        /* drop the 1st byte corresponding to RDSR command */
        spi_readbyte();

        /* read all data in the fifo */
        if ((SSP_GET(SSPSR) & 0x14))
        {
            byte = SSP_GET(SSPDR);
            if (byte & 1)
            {
                if (i % 0x20) spin_wheel();
            }
            else
                break;
        }
		udelay(100);
    }

    //udelay(1000);
    spi_rxclear();

	return (0);
}

void inline spin_wheel (void)
{
	static int p = 0;
	static char w[] = "\\/-";

	printf ("\010%c", w[p]);
	(++p == 3) ? (p = 0) : 0;
}

#endif	/* CONFIG_SPI  */
