/*
 * linux/drivers/video/ep93xxfb.c -- Frame buffer device for the EP93xx
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */
/* #define DEBUG */
#include <common.h>
#include <ep93xx.h>
#include <lcd.h>
#include "regs_raster.h"
#include "regs_syscon.h"

#define POL_HIGH				1
#define POL_LOW					0
#define EDGE_RISING				1
#define EDGE_FALLING				0
#define CLK_INTERNAL				1
#define CLK_EXTERNAL				0

vidinfo_t panel_info = {
	.vl_bpix = LCD_BPP,	/* Bits per pixel, 0 = 1, 1 = 2, 2 = 4, 3 = 8, 4 = 16 */

#if defined(LCDTYPE_QVGA)	/* 320x240@120 */
	.vl_col =	320,
	.vl_row =	240,

	.hfp =		8,
	.hsync =	48,
	.hbp =		24,
	.vfp =		4,
	.vsync =	2,
	.vbp =		16,
	.refresh =	120,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_LOW,
	.pol_vsync =	POL_LOW,
#elif defined(LCDTYPE_QVGA_LCD) /* 320x240@128 minimum blanking */
	.vl_col =	320,
	.vl_row =	240,

	.hfp =		2,
	.hsync =	2,
	.hbp =		1,
	.vfp =		0,
	.vsync =	1,
	.vbp =		0,
	.refresh =	110,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_HIGH,
	.pol_vsync =	POL_HIGH,
#elif defined(LCDTYPE_VGA) /* VGA: 640x480@60 */
	.vl_col =	640,
	.vl_row =	480,

	.hfp =		16,
	.hsync =	96,
	.hbp =		48,
	.vfp =		10,
	.vsync =	2,
	.vbp =		33,
	.refresh =	60,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_LOW,
	.pol_vsync =	POL_LOW,
#elif defined(LCDTYPE_VGA75) /* VGA: 640x480@60 */
	.vl_col =	640,
	.vl_row =	480,

	.hfp =		16,
	.hsync =	64,
	.hbp =		120,
	.vfp =		1,
	.vsync =	3,
	.vbp =		16,
	.refresh =	75,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_LOW,
	.pol_vsync =	POL_LOW,
#elif defined(LCDTYPE_SVGA) /* SVGA: 800x600@60 */
	.vl_col =	800,
	.vl_row =	600,

	.hfp =		40,
	.hsync =	128,
	.hbp =		88,
	.vfp =		1,
	.vsync =	4,
	.vbp =		23,
	.refresh =	60,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_HIGH,
	.pol_vsync =	POL_HIGH,
#elif defined(LCDTYPE_SVGA75) /* SVGA75: 800x600@75 */
	.vl_col =	800,
	.vl_row =	600,

	.hfp =		16,
	.hsync =	80,
	.hbp =		160,
	.vfp =		1,
	.vsync =	3,
	.vbp =		21,
	.refresh =	75,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_HIGH,
	.pol_vsync =	POL_HIGH,
#elif defined(LCDTYPE_XGA) /* XGA: 1024x768@60 */
	.vl_col =	1024,
	.vl_row =	768,

	.hfp =		24,
	.hsync =	136,
	.hbp =		160,
	.vfp =		3,
	.vsync =	6,
	.vbp =		29,
	.refresh =	60,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_LOW,
	.pol_vsync =	POL_LOW,
#elif defined(LCDTYPE_XGA75) /* XGA75: 1024x768@75 */
	.vl_col =	1024,
	.vl_row =	768,

	.hfp =		16,
	.hsync =	96,
	.hbp =		176,
	.vfp =		1,
	.vsync =	3,
	.vbp =		28,
	.refresh =	75,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_HIGH,
	.pol_vsync =	POL_HIGH,
#elif defined(LCDTYPE_XGAPLUS) /* XGAPLUS: 1152x864@60 */
	.vl_col =	1152,
	.vl_row =	864,

	.hfp =		32,
	.hsync =	96,
	.hbp =		192,
	.vfp =		1,
	.vsync =	3,
	.vbp =		37,
	.refresh =	60,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_HIGH,
	.pol_vsync =	POL_HIGH,
#elif defined(LCDTYPE_XGAPLUS75) /* XGAPLUS75: 1152x864@75 */
	.vl_col =	1152,
	.vl_row =	864,

	.hfp =		51,//64,
	.hsync =	102,//128,
	.hbp =		205,//256,
	.vfp =		1,
	.vsync =	2,//3,
	.vbp =		16,//32,
	.refresh =	75,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_HIGH,
	.pol_vsync =	POL_HIGH,
#elif defined(LCDTYPE_SXGA) /* SXGA: 1280x1024@60 */
	.vl_col =	1280,
	.vl_row =	1024,

	.hfp =		37,//48,
	.hsync =	86,//112,
	.hbp =		191,//248,
	.vfp =		1,
	.vsync =	2,//3
	.vbp =		19,//38,
	.refresh =	60,
	.clk_src =	CLK_INTERNAL,
	.clk_edge =	EDGE_RISING,
	.pol_blank =	POL_LOW,
	.pol_hsync =	POL_HIGH,
	.pol_vsync =	POL_HIGH,
#else
#error Please Set a LCD Type
#endif
};

static int bpp = NBITS(LCD_BPP);

int lcd_line_length;
int lcd_color_fg;
int lcd_color_bg;
void *lcd_base;		/* Start of framebuffer memory	*/
void *lcd_console_address;	/* Start of console buffer	*/
short console_col;
short console_row;
static unsigned int pseudo_palette[256];

#define ep93xxfb_outl(value, reg)				\
	{							\
		RASTER_REALITI_SWLOCK = RASTER_SWLOCK_VALUE;	\
		reg = value;				\
	}

static inline void
ep93xxfb_palette_write(ushort regno, ushort red, ushort green,
		       ushort blue)
{
    unsigned int cont, i, pal;
    pal = ((red & 0xFF) << 16) | (green & 0xFF) << 8 | ((blue & 0xFF) );
    pseudo_palette[regno] = pal;

    cont = RASTER_LUTCONT;

    if ((cont&LUTCONT_STAT && cont&LUTCONT_RAM1) ||
	(!(cont&LUTCONT_STAT) && !(cont&LUTCONT_RAM1)))
    {
       	for (i=0; i< 256; i++)  // Update inactive LUT
	{
	    (&RASTER_COLOR_LUT)[i] = pseudo_palette[i];
	}
	// Switch active LUTs next frame
	RASTER_LUTCONT = cont ^ 1;
	debug( "palette switch LUT\n" );
    } else {
      (&RASTER_COLOR_LUT)[regno] = pal;
    }
}

void
lcd_setcolreg (ushort regno, ushort red, ushort green, ushort blue)
{
	if ( NBITS(LCD_BPP) <= 8 ) {
		ep93xxfb_palette_write(regno, red, green, blue);
	}
	return;
}

static int
ep93xxfb_blank(int blank_mode)
{
	unsigned long attribs;

	attribs = RASTER_VIDEOATTRIBS & ~VIDEOATTRIBS_SDSEL_MASK;
#ifdef CONFIG_EP93XX_SDCS0
	attribs |= 0 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS1
	attribs |= 1 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS2
	attribs |= 2 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS3
	attribs |= 3 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif

	if (blank_mode) {
		ep93xxfb_outl(attribs & ~(VIDEOATTRIBS_DATAEN |
					  VIDEOATTRIBS_SYNCEN |
					  VIDEOATTRIBS_PCLKEN |
					  VIDEOATTRIBS_EN),
		     RASTER_VIDEOATTRIBS);
		RASTER_BRIGHTNESS = 0x0000;
	} else {
		if (panel_info.clk_src == CLK_INTERNAL)
			attribs |= VIDEOATTRIBS_PCLKEN;
		ep93xxfb_outl(attribs | VIDEOATTRIBS_DATAEN
			      | VIDEOATTRIBS_SYNCEN
			      | VIDEOATTRIBS_EN,
			      RASTER_VIDEOATTRIBS);
		RASTER_BRIGHTNESS = 0xFF00;
	}

	return 0;
}


static unsigned long
ep93xx_get_pll_frequency(unsigned long pll)
{
	unsigned long fb1, fb2, ipd, ps, freq;

	if (pll == 1)
		pll = SysconCLKSET1;
	else if (pll == 2)
		pll = SysconCLKSET2;
	else
		return 0;

	ps = (pll & SYSCON_CLKSET1_PLL1_PS_MASK) >>
		SYSCON_CLKSET1_PLL1_PS_SHIFT;
	fb1 = ((pll & SYSCON_CLKSET1_PLL1_X1FBD1_MASK) >>
	       SYSCON_CLKSET1_PLL1_X1FBD1_SHIFT) + 1;
	fb2 = ((pll & SYSCON_CLKSET1_PLL1_X2FBD2_MASK) >>
	       SYSCON_CLKSET1_PLL1_X2FBD2_SHIFT) + 1;
	ipd = ((pll & SYSCON_CLKSET1_PLL1_X2IPD_MASK) >>
	       SYSCON_CLKSET1_PLL1_X2IPD_SHIFT) + 1;

	freq = 225 * fb1 * fb2;

	freq = (((freq / ipd) * 65536) + (((freq % ipd) * 65536) / ipd)) >> ps;

	return freq;
}

inline long abs(long v)
{
  return (v > 0) ? v : -v;
}


static int
ep93xx_set_video_div(unsigned long freq)
{
	unsigned long f, l;
	int pdiv = 0, div = 0, psel = 0, esel = 0, i, j, k;
	long err, e;

	err = 2147483647; /* LONG_MAX */
#ifdef DEBUG
	printf( "expect=%lu\n", freq );
	printf( "Eclock=%lu\n", 14745600 );
	printf( "PLL1=%lu\n", ep93xx_get_pll_frequency(1) );
	printf( "PLL2=%lu\n", ep93xx_get_pll_frequency(2) );
#endif
	for (i = 0; i < 3 && err; i++) {
		if (i == 0)
			f = 14745600 * 2;
		else if (i == 1)
			f = ep93xx_get_pll_frequency(1) * 2;
		else
			f = ep93xx_get_pll_frequency(2) * 2;

		for (j = 4; j <= 6 && err; j++) {
			l = f / (freq * j);
			for (k = l + 1; k >= l && err; k--) {
				if (k < 2)
					break;
				e = (long)(f / (j * k)) - (long)freq;
				if (abs(e) < abs(err)) {
					pdiv = j - 3;
					div = k;
					psel = (i == 2) ? 1 : 0;
					esel = (i == 0) ? 0 : 1;
					err = e;
#ifdef DEBUG
					printf( "VCLOCK=%lu err=%ld, pdiv=%d div=%d, sel=%d\n",
						(f / (j * k)), err,
						pdiv, div, i );
#endif
				}
			}
		}
	}

	if (err == 2147483647 /*LONG_MAX*/ )
		return -1;

	SysconSWLock = 0xaa;
	SysconVIDDIV = SYSCON_VIDDIV_VENA | (esel ? SYSCON_VIDDIV_ESEL : 0) |
		(psel ? SYSCON_VIDDIV_PSEL : 0) |
		(pdiv << SYSCON_VIDDIV_PDIV_SHIFT) |
		(div << SYSCON_VIDDIV_VDIV_SHIFT);

	return freq + err;
}

static int
ep93xxfb_configure(void *base)
{
	unsigned long htotal, hfp, hsync, hbp;
	unsigned long vtotal, vfp, vsync, vbp;
	unsigned long fb_size, clk, attribs;

	fb_size = panel_info.vl_col * panel_info.vl_row * bpp / 8;

	vfp = panel_info.vfp;
	vsync = panel_info.vsync;
	vbp = panel_info.vbp;
	vtotal = panel_info.vl_row + vfp + vsync + vbp;

	hfp = panel_info.hfp;
	hsync = panel_info.hsync;
	hbp = panel_info.hbp;
	htotal = panel_info.vl_col + hfp + hsync + hbp;

	if (panel_info.clk_src == CLK_INTERNAL) {
		SysconSWLock = 0xaa;
		SysconDEVCFG &= ~SYSCON_DEVCFG_EXVC;
		clk = ep93xx_set_video_div(htotal * vtotal *
					   panel_info.refresh);
		if (clk == -1) {
			return -1;
		}

	} else {
		SysconSWLock = 0xaa;
		SysconVIDDIV = 0;
		SysconSWLock = 0xaa;
		SysconDEVCFG = SysconDEVCFG | SYSCON_DEVCFG_EXVC;
	}

	ep93xxfb_outl(0, RASTER_VIDEOATTRIBS);

	ep93xxfb_outl(vtotal - 1, RASTER_VLINESTOTAL);
	ep93xxfb_outl((vtotal - vfp - 1) + ((vtotal - vfp - vsync - 1) << 16),
		      RASTER_VSYNCSTRTSTOP);
	ep93xxfb_outl((panel_info.vl_row - 1) + ((vtotal - 1) << 16), RASTER_VACTIVESTRTSTOP);
	ep93xxfb_outl((panel_info.vl_row - 1) + ((vtotal - 1) << 16), RASTER_VBLANKSTRTSTOP);
	ep93xxfb_outl((vtotal - 1) + ((vtotal - 1) << 16), RASTER_VCLKSTRTSTOP);

	ep93xxfb_outl(htotal - 1, RASTER_HCLKSTOTAL);
	ep93xxfb_outl((htotal - 1) + ((htotal - hsync - 1) << 16),
		      RASTER_HSYNCSTRTSTOP);
	ep93xxfb_outl((htotal - hsync - hbp - 1) + ((hfp - 1) << 16),
		      RASTER_HBLANKSTRTSTOP);
	ep93xxfb_outl((htotal - hsync - hbp - 1) + ((hfp - 1) << 16),
		      RASTER_HACTIVESTRTSTOP);
	ep93xxfb_outl((htotal - 1) + ((htotal - 1) << 16), RASTER_HCLKSTRTSTOP);

	ep93xxfb_outl(0, RASTER_LINECARRY);

	RASTER_VIDSCRNPAGE = (unsigned long)base;
	RASTER_SCRNLINES = panel_info.vl_row;
	RASTER_LINELENGTH = ((panel_info.vl_col * bpp) / 32) - 1;
	RASTER_VLINESTEP = (panel_info.vl_col * bpp) / 32;

	switch (bpp)
	{
	case 4:
		RASTER_PIXELMODE = PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_4BPP |
		     PIXELMODE_C_LUT;
	  break;
	case 8:
		RASTER_PIXELMODE = PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_8BPP |
		     PIXELMODE_C_LUT;
	  break;
	case 16:
	    	RASTER_PIXELMODE = PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_16BPP |
		     PIXELMODE_C_565;
	  break;
	case 24:
	    	RASTER_PIXELMODE = PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_24BPP |
		     PIXELMODE_C_888;
	  break;
	case 32:
	    	RASTER_PIXELMODE = PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_32BPP |
		     PIXELMODE_C_888;
	  break;
	}
	SysconSWLock = 0xaa;
	SysconDEVCFG |= SYSCON_DEVCFG_RASonP3;

	attribs = 0;
#ifdef CONFIG_EP93XX_SDCS0
	attribs |= 0 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS1
	attribs |= 1 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS2
	attribs |= 2 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS3
	attribs |= 3 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
	if (panel_info.clk_edge == EDGE_FALLING)
		attribs |= VIDEOATTRIBS_INVCLK;
	if (panel_info.pol_blank == POL_HIGH)
		attribs |= VIDEOATTRIBS_BLKPOL;
	if (panel_info.pol_hsync == POL_HIGH)
		attribs |= VIDEOATTRIBS_HSPOL;
	if (panel_info.pol_vsync == POL_HIGH)
		attribs |= VIDEOATTRIBS_VCPOL;
	ep93xxfb_outl(attribs, RASTER_VIDEOATTRIBS);

	ep93xxfb_blank(0);

	return 0;
}

void lcd_ctrl_init (void *base)
{
	debug ("eplcd:lcd_ctrl_init %X\n", base);
	lcd_console_address = base;

	SysconBusMstrArb |= 0x1;

	ep93xxfb_configure(base);
}

void lcd_enable (void)
{
	debug ("eplcd:lcd_enable %X, %X %X %X\n", lcd_color_fg, lcd_color_bg,
	       lcd_console_address, lcd_console_address);
}

ulong calc_fbsize (void)
{
	ulong size;
	int line_length = (panel_info.vl_col * NBITS (panel_info.vl_bpix)) / 8;

	debug ("eplcd:calc_fbsize\n");

	size = line_length * panel_info.vl_row;
	size += PAGE_SIZE;

	return size;
}
